!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_NUCLEATION_WRAPPER
INTERFACE
SUBROUTINE ICE4_NUCLEATION_WRAPPER(KIT, KJT,KKT, LDMASK, &
                           PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                           PRVT, &
                           PCIT, PRVHENI_MR)
IMPLICIT NONE
INTEGER,                  INTENT(IN)    :: KIT, KJT, KKT
LOGICAL, DIMENSION(:,:,:),INTENT(IN)    :: LDMASK
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
END SUBROUTINE ICE4_NUCLEATION_WRAPPER
END INTERFACE
END MODULE MODI_ICE4_NUCLEATION_WRAPPER
SUBROUTINE ICE4_NUCLEATION_WRAPPER(KIT, KJT, KKT, LDMASK, &
                                   PTHT, PPABST, PRHODREF, PEXN, PLSFACT, PT, &
                                   PRVT, &
                                   PCIT, PRVHENI_MR)
!!
!!**  PURPOSE
!!    -------
!!      Computes the nucleation
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!  P. Wautelet 28/05/2019: move COUNTJV function to tools.f90
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,   ONLY: XTT

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK, ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB
use mode_tools, only: Countjv
#ifdef MNH_OPENACC
use mode_tools, only: Countjv_device
#endif

use modi_ice4_nucleation

IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
INTEGER,                  INTENT(IN)    :: KIT, KJT, KKT
LOGICAL, DIMENSION(:,:,:),INTENT(IN)    :: LDMASK
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT    ! Theta at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PPABST  ! absolute pressure at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRHODREF! Reference density
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PEXN    ! Exner function
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PT      ! Temperature at time t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRVT    ! Water vapor m.r. at t
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PCIT    ! Pristine ice n.c. at t
REAL, DIMENSION(:,:,:),   INTENT(OUT)   :: PRVHENI_MR ! Mixing ratio change due to the heterogeneous nucleation
!
!
!*       0.2  declaration of local variables
!
INTEGER                                :: IDX, JI, JJ, JK
INTEGER                                :: JL
INTEGER                                :: INEGT, INEGT_TMP
#ifndef MNH_OPENACC
INTEGER, DIMENSION(:),     ALLOCATABLE :: I1,I2,I3
LOGICAL                                :: GDSOFT
LOGICAL, DIMENSION(:),     ALLOCATABLE :: GLDCOMPUTE
LOGICAL, DIMENSION(:,:,:), ALLOCATABLE :: GNEGT  ! Test where to compute the HEN process
REAL, DIMENSION(:),        ALLOCATABLE    :: ZZT,       & ! Temperature
                                             ZPRES,      & ! Pressure
                                             ZRVT,       & ! Water vapor m.r. at t
                                             ZCIT,       & ! Pristine ice conc. at t
                                             ZTHT,       & ! Theta at t
                                             ZRHODREF,   &
                                             ZEXN,       &
                                             ZLSFACT,    &
                                             ZRVHENI_MR, &
                                             ZB_TH, ZB_RV, ZB_RI
#else
INTEGER, DIMENSION(:),     pointer, contiguous :: I1,I2,I3
LOGICAL                                        :: GDSOFT
LOGICAL, DIMENSION(:),     pointer, contiguous :: GLDCOMPUTE
LOGICAL, DIMENSION(:,:,:), pointer, contiguous :: GNEGT  ! Test where to compute the HEN process
REAL,    DIMENSION(:),     pointer, contiguous :: ZZT,       & ! Temperature
                                                  ZPRES,      & ! Pressure
                                                  ZRVT,       & ! Water vapor m.r. at t
                                                  ZCIT,       & ! Pristine ice conc. at t
                                                  ZTHT,       & ! Theta at t
                                                  ZRHODREF,   &
                                                  ZEXN,       &
                                                  ZLSFACT,    &
                                                  ZRVHENI_MR, &
                                                  ZB_TH, ZB_RV, ZB_RI
#endif
!
!$acc data present(LDMASK,PTHT,PPABST,PRHODREF,PEXN,PLSFACT,PT,PRVT,PCIT,PRVHENI_MR)

!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(LDMASK,"ICE4_NUCLEATION_WRAPPER beg:LDMASK")
  CALL MPPDB_CHECK(PTHT,"ICE4_NUCLEATION_WRAPPER beg:PTHT")
  CALL MPPDB_CHECK(PPABST,"ICE4_NUCLEATION_WRAPPER beg:PPABST")
  CALL MPPDB_CHECK(PRHODREF,"ICE4_NUCLEATION_WRAPPER beg:PRHODREF")
  CALL MPPDB_CHECK(PEXN,"ICE4_NUCLEATION_WRAPPER beg:PEXN")
  CALL MPPDB_CHECK(PLSFACT,"ICE4_NUCLEATION_WRAPPER beg:PLSFACT")
  CALL MPPDB_CHECK(PT,"ICE4_NUCLEATION_WRAPPER beg:PT")
  CALL MPPDB_CHECK(PRVT,"ICE4_NUCLEATION_WRAPPER beg:PRVT")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION_WRAPPER beg:PCIT")
END IF

#ifndef MNH_OPENACC
allocate( gnegt(kit, kjt, kkt ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_NUCLEATION_WRAPPER 1' )

CALL MNH_MEM_GET( gnegt, kit, kjt, kkt )

!$acc data present( gnegt )
#endif
!
!  optimization by looking for locations where
!  the temperature is negative only !!!
!
!$acc kernels
GNEGT(:,:,:)=PT(:,:,:)<XTT .AND. LDMASK
INEGT = COUNT(GNEGT(:,:,:))
!$acc end kernels
!
#ifndef MNH_OPENACC
ALLOCATE(GLDCOMPUTE(INEGT))
ALLOCATE(I1(INEGT),I2(INEGT),I3(INEGT))
ALLOCATE(ZZT(INEGT))
ALLOCATE(ZPRES(INEGT))
ALLOCATE(ZRVT(INEGT))
ALLOCATE(ZCIT(INEGT))
ALLOCATE(ZTHT(INEGT))
ALLOCATE(ZRHODREF(INEGT))
ALLOCATE(ZEXN(INEGT))
ALLOCATE(ZLSFACT(INEGT))
ALLOCATE(ZRVHENI_MR(INEGT))
ALLOCATE(ZB_TH(INEGT))
ALLOCATE(ZB_RV(INEGT))
ALLOCATE(ZB_RI(INEGT))
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ICE4_NUCLEATION_WRAPPER 2' )

CALL MNH_MEM_GET( GLDCOMPUTE, INEGT )
CALL MNH_MEM_GET( I1,         INEGT )
CALL MNH_MEM_GET( I2,         INEGT )
CALL MNH_MEM_GET( I3,         INEGT )
CALL MNH_MEM_GET( ZZT,        INEGT )
CALL MNH_MEM_GET( ZPRES,      INEGT )
CALL MNH_MEM_GET( ZRVT,       INEGT )
CALL MNH_MEM_GET( ZCIT,       INEGT )
CALL MNH_MEM_GET( ZTHT,       INEGT )
CALL MNH_MEM_GET( ZRHODREF,   INEGT )
CALL MNH_MEM_GET( ZEXN,       INEGT )
CALL MNH_MEM_GET( ZLSFACT,    INEGT )
CALL MNH_MEM_GET( ZRVHENI_MR, INEGT )
CALL MNH_MEM_GET( ZB_TH,      INEGT )
CALL MNH_MEM_GET( ZB_RV,      INEGT )
CALL MNH_MEM_GET( ZB_RI,      INEGT )

!$acc data present( GLDCOMPUTE, I1, I2, I3, ZZT, ZPRES, ZRVT, ZCIT, ZTHT, &
!$acc&              ZRHODREF, ZEXN, ZLSFACT, ZRVHENI_MR, ZB_TH, ZB_RV, ZB_RI )
#endif

!
#ifndef MNH_OPENACC
IF(INEGT>0) INEGT_TMP=COUNTJV(GNEGT(:,:,:), I1(:), I2(:), I3(:))
#else
IF(INEGT>0) CALL COUNTJV_DEVICE(GNEGT(:,:,:),I1(:),I2(:),I3(:),INEGT_TMP)
#endif
!
!$acc kernels
PRVHENI_MR(:,:,:)=0.
!$acc end kernels
IF(INEGT>0) THEN
!$acc kernels
!$acc loop independent
  DO JL=1, INEGT
    ZRVT(JL)=PRVT(I1(JL), I2(JL), I3(JL))
    ZCIT(JL)=PCIT(I1(JL), I2(JL), I3(JL))
    ZZT(JL)=PT(I1(JL), I2(JL), I3(JL))
    ZPRES(JL)=PPABST(I1(JL), I2(JL), I3(JL))
    ZTHT(JL)=PTHT(I1(JL), I2(JL), I3(JL))
    ZRHODREF(JL)=PRHODREF(I1(JL), I2(JL), I3(JL))
    ZEXN(JL)=PEXN(I1(JL), I2(JL), I3(JL))
    ZLSFACT(JL)=PLSFACT(I1(JL), I2(JL), I3(JL))
    GLDCOMPUTE(JL) = ZZT(JL)<XTT
  ENDDO
  GDSOFT = .FALSE.
  ZB_TH(:) = 0.
  ZB_RV(:) = 0.
  ZB_RI(:) = 0.
!$acc end kernels
  CALL ICE4_NUCLEATION(GDSOFT, GLDCOMPUTE, &
                       ZTHT, ZPRES, ZRHODREF, ZEXN, ZLSFACT, ZZT, &
                       ZRVT, &
                       ZCIT, ZRVHENI_MR, ZB_TH, ZB_RV, ZB_RI)
!$acc kernels
  PRVHENI_MR(:,:,:)= 0.0
!$acc loop independent
  DO JL=1, INEGT
    PRVHENI_MR(I1(JL), I2(JL), I3(JL)) = ZRVHENI_MR(JL)
    PCIT      (I1(JL), I2(JL), I3(JL)) = ZCIT      (JL)
  END DO
!$acc end kernels
END IF
!
IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PCIT,"ICE4_NUCLEATION_WRAPPER end:PCIT")
  !Check all OUT arrays
  CALL MPPDB_CHECK(PRVHENI_MR,"ICE4_NUCLEATION_WRAPPER end:PRVHENI_MR")
END IF

#ifdef MNH_OPENACC
!$acc end data

CALL MNH_MEM_RELEASE( 'ICE4_NUCLEATION_WRAPPER 2' )

!$acc end data

CALL MNH_MEM_RELEASE( 'ICE4_NUCLEATION_WRAPPER 1' )

!$acc end data
#endif

END SUBROUTINE ICE4_NUCLEATION_WRAPPER
