!MNH_LIC Copyright 1994-2023 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt  
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
#ifdef MNH_MGSOLVER
!     ####################
      MODULE MODI_ZSOLVER_INV
!     ####################
!
INTERFACE
!
      SUBROUTINE ZSOLVER_INV( HLBCX, HLBCY, PDXHATM, PDYHATM, PRHOM, &
                              PY, PF_1_Y )
!
!
IMPLICIT NONE
!
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
!
REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
                                                ! direction
REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
                                                ! direction
!
REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  mean of XRHODJ on the plane x y 
                                                 !  localized at a mass level
!
REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
!
REAL, DIMENSION(:,:,:), INTENT(OUT):: PF_1_Y     ! solution of the equation
!
END SUBROUTINE ZSOLVER_INV
!
END INTERFACE
!
END MODULE MODI_ZSOLVER_INV
!     ################################################################
      SUBROUTINE ZSOLVER_INV( HLBCX, HLBCY, PDXHATM, PDYHATM, PRHOM, &
                              PY, PF_1_Y )
  !     ##############################################################
  !
  !!****  *FLAT_INV * - Invert the flat quasi-laplacian operator
  !!
  !!    PURPOSE
  !!    -------
  !       This routine solves the following equation:
  !                    F ( F_1_Y ) =  Y 
  !     where F represents the quasi-laplacian without orography. The solution is 
  !     F_1_Y.
  !
  !!**  METHOD
  !!    ------
  !!      The horizontal part of F is inverted with a FFT transform. For each 
  !!    horizontal direction, the FFT form depends on the lateral boundary 
  !!    conditions :
  !!    - CRAY intrinsic function RFFTMLT in the cyclic case 
  !!    - fast cosine transform called FFT55 for all other boundary condtions.
  !!    Then, in the wavenumber space, we invert for each 
  !!    horizontal mode i,j a tridiagonal matrix by a classical double sweep 
  !!    method.  The singular mean mode (i,j)=(0,0) corresponds to the 
  !!    undetermination of the pressure to within a constant and is treated apart.
  !!    To fix this degree of freedom, we set the horizontal mean value of the 
  !!    pressure perturbation to 0 at the upper level of the model. 
  !!       
  !!    EXTERNAL
  !!    --------
  !!      Subroutine FFT55 : aplly multiple fast real staggered (shifted) 
  !!    cosine transform
  !!      Subroutine RFFTMLT : apply real-to-complex or complex-to-real Fast 
  !!    Fourier Transform (FFT) on multiple input vectors.
  !!      Subroutine FFT991  : equivalent to RFFTMLT
  !!
  !!    IMPLICIT ARGUMENTS
  !!    ------------------ 
  !!      Module MODD_PARAMETERS: declaration of parameter variables
  !!        JPHEXT, JPVEXT: define the number of marginal points out of the 
  !!        physical domain along horizontal and vertical directions respectively
  !!      Module MODD_CONF: model configurations
  !!        L2D: logical for 2D model version
  !!
  !!    REFERENCE
  !!    ---------
  !!      Book2 of documentation (subroutine FLAT_INV)     
  !!
  !!    AUTHOR
  !!    ------
  !!	P. Hereil and J. Stein       * Meteo France *
  !!
  !!    MODIFICATIONS
  !!    -------------
  !!      Original    20/07/94 
  !!      Revision  Jabouille (juillet 96) replace the CRAY intrinsic function
  !!                             RFFTMLT by the arpege routine FFT991
  !!                  17/07/97  ( J. Stein and V. Masson) initialize the corner
  !!                             verticals
  !!                  17/07/97  ( J. Stein and V. Masson) initialize the corner
  !!                             verticals
  !!      Revision  Jabouille (septembre 97) suppress the particular case for
  !!                                         tridiagonal inversion
  !!                Stein     ( January 98 ) faster computation for the unused 
  !!                                points under the ground and out of the domain  
  !!      Modification   Lugato, Guivarch (June 1998) Parallelisation
  !!                     Escobar, Stein   (July 2000) optimisation
  !-------------------------------------------------------------------------------
  !
  !*       0.    DECLARATIONS
  !              ------------
  !
  USE MODD_PARAMETERS
  USE MODD_CONF
  !
  USE MODE_ll
  USE MODD_ARGSLIST_ll, ONLY : LIST_ll
  !
  USE MODI_GET_HALO
  USE MODI_FLAT_INV
  USE MODI_DOTPROD
  !
  USE mode_mg_main_mnh
#ifdef MNH_OPENACC
  USE MODE_MNH_ZWORK,   ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
  !
  USE MODE_MPPDB
  !
  IMPLICIT NONE
  !
  !*       0.1   declarations of arguments
  !
  CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCX    ! x-direction LBC type 
  CHARACTER (LEN=4), DIMENSION(2), INTENT(IN) :: HLBCY    ! y-direction LBC type 
  !
  REAL, INTENT(IN) :: PDXHATM                     ! mean grid increment in the x
  ! direction
  REAL, INTENT(IN) :: PDYHATM                     ! mean grid increment in the y
  ! direction
  !
  REAL, DIMENSION (:), INTENT(IN) :: PRHOM         !  mean of XRHODJ on the plane x y 
  !  localized at a mass level
  !
  REAL, DIMENSION(:,:,:), INTENT(IN) :: PY         ! RHS of the equation
  !
  REAL, DIMENSION(:,:,:), INTENT(OUT):: PF_1_Y     ! solution of the equation
  !
  !*       0.2   declaration of local variables
  !
  REAL, DIMENSION(:,:,:), pointer , contiguous :: ZY ! work array to store 
  !                                                    the RHS of the equation
  INTEGER :: IIB          ! indice I for the first inner mass point along x
  INTEGER :: IIE          ! indice I for the last inner mass point along x
  INTEGER :: IJB          ! indice J for the first inner mass point along y
  INTEGER :: IJE          ! indice J for the last inner mass point along y
  INTEGER :: IKB          ! indice K for the first inner mass point along z
  INTEGER :: IKE          ! indice K for the last inner mass point along z
  INTEGER :: IKU          ! size of the arrays along z
  !
  INTEGER :: JI,JJ,JK     ! loop indexes along x, y, z respectively
  INTEGER :: IIU,IJU
  !-------------------------------------------------------------------------------
  !
  !*       1.    COMPUTE LOOP BOUNDS
  !              -------------------
  !
  if ( mppdb_initialized ) then
     !Check all in arrays
     call Mppdb_check( PRHOM,    "Zsolver_inv beg:PRHOM"       )
     call Mppdb_check( PY,       "Zsolver_inv beg:PY"       )
     call Mppdb_check( PF_1_Y,   "Zsolver_inv beg:PF_1_Y"       )
  end if
  !
  !CALL GET_PHYSICAL_ll(IIB,IJB,IIE,IJE)
  CALL GET_INDICE_ll(IIB,IJB,IIE,IJE)
  CALL GET_DIM_EXT_ll('B',IIU,IJU)
  !
  IKU=SIZE(PY,3)
  IKB=1+JPVEXT
  IKE=IKU - JPVEXT
  !
#ifndef MNH_OPENACC
  ALLOCATE(ZY(IIU,IJU,IKU))
#else
  !Pin positions in the pools of MNH memory
  CALL MNH_MEM_POSITION_PIN( 'ZSOLVER_INV' )

  CALL MNH_MEM_GET( ZY, IIU, IJU, IKU )
#endif
  !
  !-------------------------------------------------------------------------------
  !
  !*      3.    FORM HOMOGENEOUS BOUNDARY CONDITIONS FOR A NONCYCLIC CASE
  !             ---------------------------------------------------------
  !
  !
  !*      3.1 copy the RHS in a local array REMAP functions will shift the indices for the FFT
  !
  !$acc kernels
  PF_1_Y = 0.
  ZY = PY
  !$acc end kernels
  !
  !*      3.2 form homogeneous boundary condition used by the FFT for non-periodic
  !           cases
  !
  !  modify the RHS in the x direction
  !
  IF (HLBCX(1) /= 'CYCL') THEN
     !
     IF (LWEST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        DO JK=IKB,IKE                    
           DO JJ = IJB, IJE
              ZY(IIB,JJ,JK)     = ZY(IIB,JJ,JK)     + PY(IIB-1,JJ,JK)
           END DO
        END DO
        !$acc end kernels
     END IF
     !
     IF (LEAST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        DO JK=IKB,IKE
           DO JJ = IJB, IJE
              ZY(IIE,JJ,JK)     = ZY(IIE,JJ,JK)     - PY(IIE+1,JJ,JK)
           END DO
        END DO
        !$acc end kernels
     END IF
  END IF
  !
  ! modify the RHS in the same way along y
  !
  IF (HLBCY(1) /= 'CYCL'.AND. (.NOT. L2D)) THEN
     IF (LSOUTH_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        DO JK=IKB,IKE
           DO JI = IIB, IIE
              ZY(JI,IJB,JK)     = ZY(JI,IJB,JK)     + PY(JI,IJB-1,JK)
           END DO
        END DO
        !$acc end kernels
     END IF
     !
     IF (LNORTH_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        DO JK=IKB,IKE
           DO JI = IIB, IIE
              ZY(JI,IJE,JK)     = ZY(JI,IJE,JK)     - PY(JI,IJE+1,JK)
           END DO
        END DO
        !$acc end kernels
     END IF
  END IF
  !$acc wait
  !
  !-------------------------------------------------------------------------------
  !
  !*      5.    MATRIX INVERSION FOR THE FLAT OPERATOR
  !             --------------------------------------
  !
  IF (L2D) THEN
     STOP ' NO 2D MULTIGRID YET '
  END IF
  if ( mppdb_initialized ) then
     !Check all in arrays
     call Mppdb_check( ZY,     "Zsolver_inv before MG :ZY"       )     
     call Mppdb_check( PF_1_Y, "Zsolver_inv before MG :PF_1_Y"       )
  end if
  call mg_main_initialise_rhs_mnh(IIB,IIE,IIU,IJB,IJE,IJU,IKU,ZY)
  call mg_main_initialise_u_mnh(IIB,IIE,IIU,IJB,IJE,IJU,IKU,PF_1_Y)

  !print*,'************************ mg_main_mnh_solve ***************************'
       
  call mg_main_mnh_solve()  
  call mg_main_get_u_mnh(IIB,IIE,IIU,IJB,IJE,IJU,IKU,PF_1_Y)
#ifndef MNH_OPENACC
  CALL GET_HALO( PF_1_Y )
#else
  !
  !  WARNING WITH GET_HALO_D not BITREPROD !!!
  !
  CALL GET_HALO_DDC(PF_1_Y)
#endif
  !
  CALL PF_1_Y_BOUND(PF_1_Y)
  if ( mppdb_initialized ) then
     !Check all in arrays
     call Mppdb_check( PF_1_Y,   "Zsolver_inv end:PF_1_Y"       )
  end if
  !-------------------------------------------------------------------------------
#ifndef MNH_OPENACC
  DEALLOCATE(ZY)
#else
  !Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
  CALL MNH_MEM_RELEASE( 'ZSOLVER_INV' )
#endif  
  !
CONTAINS

SUBROUTINE PF_1_Y_BOUND(PF_1_Y)

IMPLICIT NONE

REAL , DIMENSION (:,:,:) :: PF_1_Y

INTEGER :: ZDXM2, ZDYM2

  !-------------------------------------------------------------------------------
  !
  !*      7.    RETURN TO A NON HOMOGENEOUS NEUMAN CONDITION FOR NON-CYCLIC CASES
  !             -----------------------------------------------------------------
  !
  !*      7.2 complete the lateral boundaries
  !
  IF (HLBCX(1) /= 'CYCL') THEN
     !
     !*      7.2.1 return to a non-homogeneous case in the x direction
     ! 
     ZDXM2 = PDXHATM*PDXHATM
     !
     IF (LWEST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        !$acc loop independent collapse(2)
        DO JK=IKB,IKE
           DO JJ = IJB,IJE
              PF_1_Y(IIB-1,JJ,JK) = PF_1_Y(IIB,JJ,JK) - PY(IIB-1,JJ,JK)*ZDXM2/PRHOM(JK)
           END DO
        END DO
        !$acc end kernels
     END IF
     !
     IF (LEAST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        !$acc loop independent collapse(2)
        DO JK=IKB,IKE
           DO JJ = IJB,IJE
              PF_1_Y(IIE+1,JJ,JK) = PF_1_Y(IIE,JJ,JK) + PY(IIE+1,JJ,JK)*ZDXM2/PRHOM(JK)
           END DO
        END DO
        !$acc end kernels
     END IF
     !$acc wait
     !
     !    we set the solution at the corner point by the condition:
     !    dxm ( P ) = 0
     IF (LWEST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        !$acc loop independent
        DO JJ = IJB,IJE
           PF_1_Y(IIB-1,JJ,IKB-1) = PF_1_Y(IIB,JJ,IKB-1) 
           PF_1_Y(IIB-1,JJ,IKE+1) = PF_1_Y(IIB,JJ,IKE+1)
        END DO
        !$acc end kernels
     END IF
     IF (LEAST_ll(HSPLITTING='B')) THEN
        !$acc kernels async
        !$acc loop independent
        DO JJ = IJB,IJE
           PF_1_Y(IIE+1,JJ,IKB-1) = PF_1_Y(IIE,JJ,IKB-1)
           PF_1_Y(IIE+1,JJ,IKE+1) = PF_1_Y(IIE,JJ,IKE+1)
        END DO
        !$acc end kernels
     END IF
     !$acc wait
     !
  ELSE
     !
     !*      7.2.2 periodize the pressure function field along the x direction
     !
     ! in fact this part is useless because it is done in the routine
     ! REMAP_X_2WAY.
     !
  END IF
!!$!
  IF (.NOT.L2D) THEN
     IF (HLBCY(1) /= 'CYCL') THEN
        !
        !*      7.2.3 return to a non-homogeneous case in the y direction
        !
        ZDYM2 = PDYHATM*PDYHATM 
        !
        IF (LSOUTH_ll(HSPLITTING='B')) THEN
           !$acc kernels async
           !$acc loop independent collapse(2)
           DO JK=IKB,IKE
              DO JI = IIB,IIE
                 PF_1_Y(JI,IJB-1,JK) = PF_1_Y(JI,IJB,JK) - PY(JI,IJB-1,JK)*ZDYM2/PRHOM(JK)
              END DO
           END DO
           !$acc end kernels
        END IF
        !
        IF (LNORTH_ll(HSPLITTING='B')) THEN
           !$acc kernels async
           !$acc loop independent collapse(2)
           DO JK=IKB,IKE
              DO JI = IIB,IIE
                 PF_1_Y(JI,IJE+1,JK) = PF_1_Y(JI,IJE,JK) + PY(JI,IJE+1,JK)*ZDYM2/PRHOM(JK)
              END DO
           END DO
           !$acc end kernels
        END IF
        !$acc wait
        !    we set the solution at the corner point by the condition:
        !    dym ( P )  = 0
        !
        IF (LSOUTH_ll(HSPLITTING='B')) THEN
           !$acc kernels async
           !$acc loop independent
           DO JI = IIB,IIE
              PF_1_Y(JI,IJB-1,IKB-1) = PF_1_Y(JI,IJB,IKB-1)
              PF_1_Y(JI,IJB-1,IKE+1) = PF_1_Y(JI,IJB,IKE+1)
           END DO
           !$acc end kernels
        END IF
        !
        IF (LNORTH_ll(HSPLITTING='B')) THEN
           !$acc kernels async
           !$acc loop independent
           DO JI = IIB,IIE
              PF_1_Y(JI,IJE+1,IKB-1) = PF_1_Y(JI,IJE,IKB-1)
              PF_1_Y(JI,IJE+1,IKE+1) = PF_1_Y(JI,IJE,IKE+1)
           END DO
           !$acc end kernels
        END IF
        !$acc wait
     ELSE
        !
        !*      7.2.4 periodize the pressure function field along the y direction
        !
        !
        ! in fact this part is useless because it is done in the routine
        ! REMAP_X_2WAY.
        !
     END IF
     !
  END IF
  !
  IF (.NOT. L2D .AND. HLBCX(1)/='CYCL' .AND. HLBCY(1)/='CYCL') THEN
     ! the following verticals are not used
     IF ( (LWEST_ll(HSPLITTING='B')).AND.(LSOUTH_ll(HSPLITTING='B')) ) THEN
        !$acc kernels async
        PF_1_Y(IIB-1,IJB-1,:)= PF_1_Y(IIB,IJB,:) ! 0.0
        !$acc end kernels
     END IF
     !
     IF ( (LWEST_ll(HSPLITTING='B')).AND.(LNORTH_ll(HSPLITTING='B')) ) THEN
        !$acc kernels async
        PF_1_Y(IIB-1,IJE+1,:)= PF_1_Y(IIB,IJE,:) ! 0.0
        !$acc end kernels
     END IF
     !
     IF ( (LEAST_ll(HSPLITTING='B')).AND.(LSOUTH_ll(HSPLITTING='B')) ) THEN
        !$acc kernels async
        PF_1_Y(IIE+1,IJB-1,:)= PF_1_Y(IIE,IJB,:) ! 0.0
        !$acc end kernels
     END IF
     !
     IF ( (LEAST_ll(HSPLITTING='B')).AND.(LNORTH_ll(HSPLITTING='B')) ) THEN
        !$acc kernels async
        PF_1_Y(IIE+1,IJE+1,:)= PF_1_Y(IIE,IJE,:) ! 0.0
        !$acc end kernels
     END IF
     !$acc wait
  END IF
END SUBROUTINE PF_1_Y_BOUND
  !------------------------------------------------------------------------------  
END SUBROUTINE ZSOLVER_INV
#endif
